// todo: how to set samplerate for new DynMultiEQ?
// amp: can't have amp sent to EVERY band in DynMultiEQ

DynMultiEQ {
	// container for specifications for an EQ bank
	// creates an individual synthdef for each band and allows placing on server
	// H. James Harkins, Jan 2003 - jamshark70@yahoo.com
	// requires Joseph Anderson's Regalia-Mitra EQ classes
	// ftp://ftp.music.washington.edu/Public/jla/RM_Filters.sit
	
	classvar	numArgs = 4,	// must equal the number of instance variables below
			<>sampleRate = 44100;
	
	var	<spec,	// List of EQBand
		<numChannels,
		<target, <bus,
		<isPlaying = false;
	
	*new {
			  // takes an argument list; .init calls .parseArgs to parse the list
		arg numChan = 1 ... z;
		
		^super.new.init(numChan, z)
	}
	
	init { arg numChan ... z;
		numChannels = numChan;
		spec = this.parseArgs(z);
	}
	
	parseArgs {  // turn argument list into array of parameter arrays
		arg ... z;	// argument list; may or may not be wrapped in an array
		
		var oneSpec, allSpecs;

		z = z.flat;	// remove all inner array wrapping

		(z.size == 0).if({
			^List.new		// empty list for no specs
		}, {
			allSpecs = List.new;		// initialize allSpecs
			
			z.do({ arg item;	// iterate over arguments
				item.isKindOf(Symbol).if( {	// a symbol indicates a new spec
						// if oneSpec is nil, this is the first one so don't add
					oneSpec.isNil.if(nil,
						{ 		// else add it
							allSpecs.add(EQBand.newFromArray(oneSpec, numChannels, sampleRate));
						}
					);
					oneSpec = List.with(item);	// then initialize new spec
				}, {
					oneSpec.add(item);		// other types, add to onespec
				});
			});
			
			^allSpecs.add(EQBand.newFromArray(oneSpec, numChannels, sampleRate));  // add the last spec
	
		});
	}
	
	add {		// add one or more filters
		arg ... z;		// same syntax for arglist as .new
		var newbands;
		
		newbands = this.parseArgs(z);
		isPlaying.if({
			newbands.do({ arg b;
				b.play(target, bus/*, addAction, amp*/);
			});
		});
		spec.add(newbands);
		
		^this
	}
	
	removeAt {		// remove a filter
		arg i;	// index to filter
		var band;
		
		band = spec.removeAt(i);
		band.free;

		^this
	}
	
	at {		// index a filter; returns one band
		arg i;
		^spec.at(i)
	}
	
	play { arg targ, b, addAction = \addToTail, amp = 1;
		target = targ;
		bus = b;
		spec.do({ arg sp, i;
			sp.play(target, bus, addAction, (i == 0).if({ amp }, { 1 }));
		});
	}
	
	free {
		spec.do({ arg sp; sp.free });
	}
	
// global setters? probably not useful, but add them anyway

	asString { 
		var out;
		out = "MultiEQ.new(" ++ numChannels ++ ", ";   // initialize string
		spec.do( {		// loop through filters
			arg s, i;
			(i > 0).if({ out = out ++ ", " });  // skip comma for first eq
			out = out ++ "\\" ++ s.type
				++ ", " ++ s.freq
				++ ", " ++ s.k
				++ ", " ++ s.rq;
		});
		^(out ++ ")")
	}
	
	post {	// add post and postln support
		this.asString.post;
		^this
	}
	
	postln {
		this.asString.postln;
		^this
	}

		// can't use this as is with the server	
//	arAt { 	// index a single filter and send result
//			// to use several filters at once, feed an EQspecs to a MultiEQ.ar
//		arg in, i;		// input signal, index to filter
//		
//		var filters;
//		
//		filters = MultiEQ.filterTypes;
//		
//		^filters.at(ftypes.at(i)).value(in, freqs.at(i), gains.at(i), rqs.at(i))
//	}

//	*tune {	// open a GUI to find settings for something playing on a given bus
//		arg	target, bus,
//			numChan = 2,
//			mul = 1
//			... specs;		// takes a ugen and initial specifications as input
//		
//		var initSpec, gui;
//		
//		specs = specs.flat;
//		specs.at(0).isKindOf(EQspecs).if( {	// if specs are already compiled
//			initSpec = specs.at(0);		// use specs, ignore other arguments
//		}, {
//			initSpec = EQspecs.new(specs);	// create specs from argument list
//		});
//		
//			// how will this work?
////		gui = EQspecGUI.new(initSpec).play(ugenFunc, numChan, mul);  // open GUI and play function
//
//		gui.spec.postln;		// print new spec
//
//		// output an EQspecs object
//		^gui.spec
//	}

}

MultiEQ : DynMultiEQ { }

StaticEQ {
	// make a static synthdef out of given specs
	
	classvar	<eqFuncs,
//			<>dir,
			<>sampleRate = 44100;
	
	var	<spec,
		<defname,
		<numChannels,
		<target, <bus,
		<mul,		// volume scaling after eq
		<synth,
//		<addAction = \addToTail,
		<readyForPlay = false;
	
	*initClass {
			// each func takes "in" and outputs eq'ed signal (1 band)
		eqFuncs = IdentityDictionary[
			\eq -> { arg in, k, a0, a1;
				var allpass;
				allpass = SOS.ar(in, a0, a1, 1, a1.neg, a0.neg, k.sign);
				0.5 * (in + allpass + (k.abs * (in-allpass)))
			},
			\loshelf -> { arg in, k, b1;
				var allpass;
				allpass = FOS.ar(in, b1.neg, 1, b1, k.sign);
				0.5 * (in + allpass + (k.abs * (in-allpass)))
			},
			\hishelf -> { arg in, k, b1;
				var allpass;
				allpass = FOS.ar(in, b1.neg, 1, b1, k.sign);
				0.5 * (in + allpass + (k.abs * (in-allpass)))
			}
		];
	}
	
	*new {
		arg numChan = 1 ... z;
		^super.new.init(numChan, z)
	}
	
	init { arg numChan ... z;
		numChannels = numChan ? 1;
		spec = DynMultiEQ.new(numChan, z).spec;
	}
	
	prepareForPlay { arg targ, b, addAction = \addToTail;
//"StaticEQ-prepareForPlay".postln;

		this.fixTarget(targ, b, addAction);
			// will this choke if completionMsg is nil?
		this.asSynthDef.send(target.server);
		
		readyForPlay = true;
	}
	
	play { arg targ, b, addAction = \addToTail, amp = 1;
			// no provision for changing synth--
			// if you need to change, use DynMultiEQ
		var bundle, delay = nil;  // delay is needed if making the synthdef right before play
//"StaticEQ-play".postln;

				// if already playing and not moving, do nothing
		(synth.notNil && targ.isNil).if({ ^this });
				// now we know targ is not nil, so we're moving
		this.fixTarget(targ, b, addAction);
		synth.notNil.if({
			synth.perform(EQBand.translateMethods.at(addAction), target);
			synth.set(\bus, bus.index);
			^this	// go back
		});

			// making a new synth
		bundle = List.new;
		mul = amp ? mul ? 1;

			// if it's ready for play now, send the s_new message
		readyForPlay.if({
			synth = Synth.newToBundle(bundle, defname, 
				[\bus, bus.index, \mul, mul], target, addAction);
			target.server.listSendBundle(nil, bundle);
		}, {		// if it's not ready, make the synthdef, send it, and create the synth when done
			defname = this.makeSynthName;
			synth = Synth.newToBundle(bundle, defname, 
				[\bus, bus.index, \mul, mul], target, addAction);
			this.asSynthDef.send(target.server, bundle.at(0));
		});
		
	}
	
	mul_ { arg amp;
		mul = amp ? mul ? 1;
		synth.notNil.if({ synth.set(\mul, mul) });
	}
	
	free {
		synth.notNil.if({ synth.free; });
		synth = nil;
	}
	
	move { arg targ, moveAction = \moveAfter;
		synth.notNil.if({
			target = targ.asTarget;
			synth.perform(moveAction, target);
		});
	}

	fixTarget { arg targ, b, addAction = \addToTail;
	
		target = targ ? target;
		bus = b;

		target.isKindOf(Group).not.if({	// target is Group, leave it alone, otherwise...
			target.isKindOf(Synth).if({
				addAction = \addAfter;	// if a synth, place imm. after target
			});
			target.isKindOf(MixerChannel).if({	// if mixerchannel,
				bus = target.inbus;	// must use mc bus
				target = target.effectgroup;	// place where effects should go
			});
			target = target.asTarget;
		});
		
		bus.isKindOf(Integer).if({	// send an index, make a bus from it
			bus = Bus.new(\audio, bus, numChannels, target.server);
		});
		
		bus.isNil.if({	// give nothing for bus, use hardware on target's server
			bus = Bus.new(\audio, 0, numChannels, target.server);
		});
	}	
	
	asSynthDef { arg numChan = 1;
		defname = defname ? this.makeSynthName;
		^SynthDef.new(this.defname, {
			arg bus, mul;
			var sig;
			sig = In.ar(bus, numChannels);
			spec.do({ arg sp;
				sig = eqFuncs.at(sp.type).value(sig, sp.k, sp.a0, sp.a1);
			});
			ReplaceOut.ar(bus, sig * mul);
		});
	}
	
	makeSynthName {
		var str;
		str = Main.elapsedTime.trunc(0.001).asString;
		str.do({ arg s, i; (s == $.).if({ str.put(i, $-) }); });
		^("EQ" ++ str)
	}
	
}